# Copyright 2023 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
from jax import core
import itertools
import weakref
from functools import lru_cache, wraps


def weak_lru(maxsize=128, typed=False):
  'LRU Cache decorator that keeps a weak reference to "self"'

  def wrapper(func):

    @lru_cache(maxsize, typed)
    def _func(_self, *args, **kwargs):
      return func(_self(), *args, **kwargs)

    @wraps(func)
    def inner(self, *args, **kwargs):
      return _func(weakref.ref(self), *args, **kwargs)

    return inner

  return wrapper


def complement_argnums(argnums, num_args):
  argnums = (argnums,) if isinstance(argnums, int) else argnums
  complement = []
  a = 0
  for i in range(num_args):
    if a < len(argnums) and i == argnums[a]:
      a += 1
    else:
      complement.append(i)
  return tuple(complement)


# visualization
# adopted from
# https://gist.github.com/niklasschmitz/559a1f717f3535db0e26d0edccad0b46

styles = {
  'const': dict(style='filled', color='goldenrod1'),
  'invar': dict(color='mediumspringgreen', style='filled'),
  'outvar': dict(style='filled,dashed', fillcolor='indianred1', color='black'),
  'op_node': dict(shape='box', color='lightskyblue', style='filled'),
  'intermediate': dict(style='filled', color='cornflowerblue')
}


def _jaxpr_graph(jaxpr):
  from graphviz import Digraph
  id_names = (f'id{id}' for id in itertools.count())
  graph = Digraph(engine='dot')
  graph.attr(size='6,10!')
  for v in jaxpr.constvars:
    graph.node(
      str(v),
      core.raise_to_shaped(v.aval).str_short(), styles['const']
    )
  for v in jaxpr.invars:
    graph.node(str(v), v.aval.str_short(), styles['invar'])
  for eqn in jaxpr.eqns:
    for v in eqn.invars:
      if isinstance(v, core.Literal):
        graph.node(
          str(id(v.val)),
          core.raise_to_shaped(core.get_aval(v.val)).str_short(),
          styles['const']
        )
    if eqn.primitive.multiple_results:
      id_name = next(id_names)
      graph.node(id_name, str(eqn.primitive), styles['op_node'])
      for v in eqn.invars:
        graph.edge(
          str(id(v.val) if isinstance(v, core.Literal) else v), id_name
        )
      for v in eqn.outvars:
        graph.node(str(v), v.aval.str_short(), styles['intermediate'])
        graph.edge(id_name, str(v))
    else:
      outv, = eqn.outvars
      graph.node(str(outv), str(eqn.primitive), styles['op_node'])
      for v in eqn.invars:
        graph.edge(
          str(id(v.val) if isinstance(v, core.Literal) else v), str(outv)
        )
  for i, v in enumerate(jaxpr.outvars):
    outv = 'out_' + str(i)
    graph.node(outv, outv, styles['outvar'])
    graph.edge(str(v), outv)
  return graph


def visualize_jaxpr(fn):

  def wrapper(*args, **kwargs):
    jaxpr = jax.make_jaxpr(fn)(*args, **kwargs).jaxpr
    g = _jaxpr_graph(jaxpr)
    filename = str(hash(jaxpr))
    g.view(filename=filename, directory="/tmp", cleanup=True)
    print(filename)
    return fn(*args, **kwargs)

  return wrapper



# Copyright 2023 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Core functions."""
import jax
import jax.numpy as jnp
import jaxtyping
from jax.tree_util import (
  tree_flatten,
  tree_map,
  tree_unflatten,
  tree_structure,
)
import jax.extend.linear_util as lu
from typing import Tuple
import inspect
import types
import numpy as np
from dataclasses import dataclass
from .utils import weak_lru
from functools import partial

# Make Tracer hashable
jax.core.Tracer.__hash__ = lambda self: id(self)

# this is to trick jax to return dtype=float32 for functions
# TODO: generalize dtype for general arrays (aka functions).
MetaAbstractArray = jaxtyping._array_types._MetaAbstractArray  # noqa

dtype_to_jaxtyping = {
  "uint8": jaxtyping.UInt8,
  "uint16": jaxtyping.UInt16,
  "uint32": jaxtyping.UInt32,
  "uint64": jaxtyping.UInt64,
  "int8": jaxtyping.Int8,
  "int16": jaxtyping.Int16,
  "int32": jaxtyping.Int32,
  "int64": jaxtyping.Int64,
  "bfloat16": jaxtyping.BFloat16,
  "float16": jaxtyping.Float16,
  "float32": jaxtyping.Float32,
  "float64": jaxtyping.Float64,
  "complex64": jaxtyping.Complex64,
  "complex128": jaxtyping.Complex128,
  "bool": jaxtyping.Bool,
  "bool_": jaxtyping.Bool,
}


class Spec:

  def __init__(self, shape, dtype):
    self.shape = shape
    self.dtype = np.dtype(dtype)

  def __repr__(self) -> str:
    return f"{self.dtype.char}{list(self.shape)}"

  def __eq__(self, other):
    if isinstance(other, Spec):
      return self.shape == other.shape and self.dtype == other.dtype
    else:
      return False

  @staticmethod
  def from_annotation(ann):
    if isinstance(ann, MetaAbstractArray):
      if len(ann.dtypes) > 1:
        raise ValueError(
          f"Currently only support single dtype, got {ann}, "
          f"which has multiple dtypes of {ann.dtypes}."
        )
      return Spec(tuple(_normalize_dim(dim) for dim in ann.dims), ann.dtypes[0])
    else:
      raise ValueError(f"Unknown annotation: {ann} of type {type(ann)}")


class SpecTree:
  """Utility functions for converting between Pytree[Spec] and Annotation.
  """

  @staticmethod
  def from_annotation(ann):
    if hasattr(ann, "_name") and ann._name == "Tuple":
      return tuple(SpecTree.from_annotation(a) for a in ann.__args__)
    else:
      return Spec.from_annotation(ann)

  @classmethod
  def from_args(cls, f):
    if hasattr(f, "arg_spec"):
      return f.arg_spec
    elif callable(f):
      return tuple(
        cls.from_annotation(p.annotation)
        for p in inspect.signature(f).parameters.values()
      )
    elif hasattr(f, "aval") and isinstance(f.aval, GeneralArray):
      return f.aval.arg_spec
    elif hasattr(f, "shape") and isinstance(f.shape[0], GeneralDim):
      return tuple(d.spec for d in f.shape[1:])
    else:
      raise ValueError(f"Invalid function: {f} of type {type(f)}")

  @classmethod
  def from_ret(cls, f):
    if hasattr(f, "ret_spec"):
      return f.ret_spec
    elif callable(f):
      return cls.from_annotation(inspect.signature(f).return_annotation)
    elif hasattr(f, "aval") and isinstance(f.aval, GeneralArray):
      return f.aval.ret_spec
    elif hasattr(f, "shape") and isinstance(f.shape[0], GeneralDim):
      return f.shape[0].spec
    else:
      raise ValueError(f"Invalid function: {f} of type {type(f)}")

  @classmethod
  def from_value(cls, v):
    if isinstance(v, float):
      return Spec((), jnp.float32)
    return tree_map(lambda x: Spec(x.shape, x.dtype), v)


class GeneralArray(jax.core.ShapedArray):

  def __init__(self, shape, dtype=jnp.float32, weak_type=True):
    # TODO: currently setting dtype to types.FunctionType will cause
    # weird error.
    super().__init__(shape, dtype=dtype, weak_type=weak_type)

  def update(self, shape=None, dtype=None, weak_type=None):
    sa = super().update(shape, dtype, weak_type)
    return GeneralArray(sa.shape, sa.dtype, sa.weak_type)

  def at_least_vspace(self):
    sa = super().at_least_vspace()
    return GeneralArray(sa.shape, sa.dtype, sa.weak_type)

  def str_short(self, short_dtypes=True):
    ret = self.shape[0]
    arg_str = ",".join(
      map(lambda x: "{...}" if isinstance(x, dict) else str(x), self.arg_spec)
    )
    return f"{arg_str}->{ret}".replace(" ", "")

  @property
  def num_args(self):
    return len(self.shape) - 1

  @property
  def arg_spec(self):
    return tuple(d.spec for d in self.arg)

  @property
  def ret_spec(self):
    return self.ret.spec

  @property
  def arg(self):
    return self.shape[1:]

  @property
  def ret(self):
    return self.shape[0]

  def to_tangent_aval(self):
    return GeneralArray(self.shape, self.dtype, self.weak_type)


jax.core.raise_to_shaped_mappings[GeneralArray
                                 ] = lambda aval, weak_type: GeneralArray(
                                   aval.shape,
                                   dtype=aval.dtype,
                                   weak_type=weak_type,
                                 )


class operator:
  """Decorating a WrappedFun with this class makes it cached.
  After this decoration, no more wrapper can be appended to the transforms
  stack of the original WrappedFun f.

  When applying lu.transformations on this class, a new WrappedFun is created
  with this class as the core function.
  """

  def __init__(self, f):
    assert isinstance(f, lu.WrappedFun)
    self.f = f

  @weak_lru(maxsize=128, typed=False)
  def call_flatten_args_cached(self, tree_def, *args):
    args, kwargs = tree_unflatten(tree_def, args)
    return self.f.call_wrapped(*args, **kwargs)

  def __call__(self, *args, **kwargs):
    flat_args, tree_def = tree_flatten((args, kwargs))
    if all(map(lambda a: isinstance(a, jax.core.Trace), flat_args)):
      return self.call_flatten_args_cached(tree_def, *flat_args)
    else:
      return self.f.call_wrapped(*args, **kwargs)

  def wrap(self, gen, gen_static_args, out_store):
    """Add another transform and its store."""
    if out_store is None:
      return lu.WrappedFun(
        self, partial(gen, self, *gen_static_args), ((gen, gen_static_args),),
        (out_store,), (), None, None
      )
    else:
      return lu.WrappedFun(
        self, partial(gen, self, out_store, *gen_static_args),
        ((gen, gen_static_args),), (out_store,), (), None, None
      )

  def __repr__(self):
    return f"{self.f}"

  def __hash__(self):
    return hash(self.f)


class function:
  """Wrapper for a function, to add some operator overloading.
  """

  def __init__(self, f):
    if not isinstance(f, lu.WrappedFun):
      self.shape = (Ret(SpecTree.from_ret(f)), *map(Arg, SpecTree.from_args(f)))
      f = lu.wrap_init(f)
    self.f = f
    self._f_hash = hash(self.f)
    self._jaxpr = None
    self._arrays = {}

  @property
  def grid(self):
    if "grid" not in self._arrays:
      self._arrays["grid"] = Grid.none(len(self.arg))
    return self._arrays["grid"]

  @grid.setter
  def grid(self, g):
    self._arrays["grid"] = g

  def wrap(self, gen, gen_static_args, out_store):
    """Add another transform and its store."""
    if out_store is None:
      return lu.WrappedFun(
        self, partial(gen, self, *gen_static_args), ((gen, gen_static_args),),
        (out_store,), (), None, None
      )
    else:
      return lu.WrappedFun(
        self, partial(gen, self, out_store, *gen_static_args),
        ((gen, gen_static_args),), (out_store,), (), None, None
      )

  def __repr__(self) -> str:
    arg_str = ",".join(
      map(lambda x: "{...}" if isinstance(x, dict) else str(x), self.arg_spec)
    )
    return f"{arg_str} -> {self.ret_spec}"

  def __hash__(self):
    # as long as function call cache is concerned,
    # the hash is independent on self._arrays.
    return self._f_hash

  @weak_lru(maxsize=128, typed=False)
  def call_flatten_cached(self, tree_def, *args):
    return tree_flatten(self.f.call_wrapped(*tree_unflatten(tree_def, args)))[0]

  def _call(self, *args):
    out_def = tree_structure(self.ret_spec)
    flat_args, tree_def = tree_flatten(args)
    flat_args = tuple(flat_args)
    if all(map(lambda a: isinstance(a, jax.core.Tracer), flat_args)):
      return tree_unflatten(
        out_def, self.call_flatten_cached(tree_def, *flat_args)
      )
    else:
      # if evaluated with ArrayImpl
      if self._jaxpr is None:
        self._jaxpr = jax.make_jaxpr(
          lambda *args: self.call_flatten_cached(tree_def, *args)
        )(*flat_args)
        self._jaxpr_fun = jax.core.jaxpr_as_fun(self._jaxpr)
      return tree_unflatten(out_def, self._jaxpr_fun(*flat_args))

  @property
  def num_args(self):
    return len(self.shape) - 1

  @property
  def arg_spec(self):
    return tuple(a.spec for a in self.shape[1:])

  @property
  def ret_spec(self):
    return self.ret.spec

  @property
  def arg(self):
    return self.shape[1:]

  @property
  def ret(self):
    return self.shape[0]


def with_spec(arg_spec=None, ret_spec=None):

  def decorator(f):
    ff = function(lu.wrap_init(f))
    ff.shape = (Ret(ret_spec), *map(Arg, arg_spec))
    return ff

  return decorator


# Extend the original dimension system with GeneralDim,
# to denote the inputs/output of functions.
class GeneralDim:
  """GeneralDim class, denotes either the input / output
  of a jaxtyping annotated python function.
  """

  def __init__(self, spec):
    self.spec = spec

  def __eq__(self, other):
    if isinstance(other, GeneralDim):
      return self.spec == other.spec
    else:
      return False

  def __gt__(self, other):
    return True

  def __ge__(self, other):
    return True

  def __repr__(self) -> str:
    return repr(self.spec)

  def __hash__(self) -> str:
    return hash(repr(self.spec))

  def evaluate(self, env):
    return self

  def dimension_as_value(self) -> int:
    pass

  def to_shaped_array(self):
    return tree_map(
      lambda s: jax.core.ShapedArray(s.shape, dtype=s.dtype), self.spec
    )


def Arg(spec):
  return GeneralDim(spec)


def Ret(spec):
  return GeneralDim(spec)


def num_args(f):
  return getattr(f, "num_args", len(SpecTree.from_args(f)))


def ret(f):
  return Ret(SpecTree.from_ret(f))


def ret_like(v):
  return Ret(SpecTree.from_value(v))


# Utility functions for interoperation between
# jaxtyping, shape in pytree, and GeneralDim.
def _normalize_dim(dim):
  if isinstance(dim, (GeneralDim, int)):
    return dim
  elif isinstance(dim, jaxtyping._array_types._SymbolicDim):  # noqa
    return _normalize_dim(eval(dim))
  elif isinstance(dim, jaxtyping._array_types._FixedDim):  # noqa
    return dim.size
  else:
    raise ValueError(f"Unknown dim type: {dim} of type {type(dim)}")


# Convert a function to its abstract value.
def function_to_aval(f):
  """Convert a function to its abstract value."""
  if isinstance(f, types.FunctionType):
    raise ValueError(
      f"Please decorate the function {f} with @function or @with_spec."
    )
  elif isinstance(f, function):
    return GeneralArray(f.shape)


def is_function(f):
  # identify whether it is tracer first
  if hasattr(f, "aval"):
    return isinstance(f.aval, GeneralArray)
  elif callable(f):
    return True
  elif isinstance(f, GeneralArray):
    return True


def zeros_like(f):

  @with_spec(arg_spec=SpecTree.from_args(f), ret_spec=SpecTree.from_ret(f))
  def _zero(*args):
    return tree_map(jnp.zeros, SpecTree.from_ret(f))

  return _zero


def dummy_input(f):
  return dummy_array(SpecTree.from_args(f))


def dummy_output(f):
  return dummy_array(SpecTree.from_ret(f))


def random_input(key, f):
  return tree_map(
    lambda s: jax.random.normal(key, s.shape, dtype=s.dtype),
    SpecTree.from_args(f)
  )


def dummy_array(spec_tree):
  return tree_map(
    lambda spec: jnp.zeros(spec.shape, dtype=spec.dtype),
    spec_tree,
  )


def jacobian_spec(f, argnums=0, has_aux=False):
  i_spec = SpecTree.from_args(f)
  o_spec = SpecTree.from_ret(f)
  if has_aux:
    if not isinstance(o_spec, tuple) or len(o_spec) != 2:
      raise ValueError("has_aux=True requires f to have 2 outputs.")
    o_spec, aux_spec = o_spec
  else:
    aux_spec = None
  o_flat, o_def = tree_flatten(o_spec)
  if isinstance(argnums, int):
    i_spec = i_spec[argnums]
  elif isinstance(argnums, (tuple, list)):
    i_spec = tuple(i_spec[i] for i in argnums)
  i_flat, i_def = tree_flatten(i_spec)
  jac_spec = tree_unflatten(
    o_def,
    list(
      tree_unflatten(
        i_def, list(Spec(o.shape + i.shape, i.dtype) for i in i_flat)
      ) for o in o_flat
    )
  )
  if has_aux:
    return jac_spec, aux_spec
  else:
    return jac_spec


@dataclass
class Grid:
  nodes: Tuple[jnp.ndarray]
  weights: Tuple[jnp.ndarray]

  @staticmethod
  def none(num):
    return Grid((None,) * num, (None,) * num)

  def take(self, indice):
    return Grid(
      tuple(map(self.nodes.__getitem__, indice)),
      tuple(map(self.weights.__getitem__, indice)),
    )

  def __add__(self, other):
    return Grid(self.nodes + other.nodes, self.weights + other.weights)

  def __radd__(self, other):
    if other == 0 or other is None:
      return self
    else:
      return other + self


# register with jax, so that when a python function is used for tracing,
# it is converted to an abstract ShapedArray type, that has infinite dimensions.
jax.core.pytype_aval_mappings[types.FunctionType] = function_to_aval
jax.interpreters.xla.pytype_aval_mappings[types.FunctionType] = function_to_aval
jax.interpreters.xla.canonicalize_dtype_handlers[types.FunctionType
                                                ] = lambda x: x
jax._src.api_util._shaped_abstractify_handlers[types.FunctionType
                                              ] = function_to_aval
jax._src.dtypes.python_scalar_dtypes[types.FunctionType] = np.dtype("float32")

# register function class
jax.core.pytype_aval_mappings[function] = function_to_aval
jax.interpreters.xla.pytype_aval_mappings[function] = function_to_aval
jax.interpreters.xla.canonicalize_dtype_handlers[function] = lambda x: x
jax._src.api_util._shaped_abstractify_handlers[function] = function_to_aval
jax._src.dtypes.python_scalar_dtypes[function] = np.dtype("float32")



# Copyright 2023 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A collection of operators."""

import types
import numpy as np
import functools
from types import SimpleNamespace
import jax
import jax.extend.linear_util as lu
import jax.numpy as jnp
from jax.tree_util import (
  tree_map,
  tree_leaves,
  tree_structure,
)
from jax.interpreters import ad
from jax._src import ad_util
from jax import core
from jax._src.util import HashableFunction
from absl import logging
from autofd.utils import complement_argnums
from . import transforms as t
from autofd.general_array import (
  ret_like,
  Ret,
  Arg,
  Spec,
  SpecTree,
  GeneralArray,
  function,
  operator,
  with_spec,
  num_args,
  jacobian_spec,
  dummy_array,
  is_function,
  Grid,
)


def def_operator(prim, impl, spec, grid, multiple_results=False):
  prim.multiple_results = multiple_results

  def _set_shape_and_dtype(v, shape, dtype):
    if isinstance(v, function):
      v.shape = shape

  def _set_grid(v, grid):
    if isinstance(v, function):
      v.grid = grid

  @prim.def_impl
  def _impl(*args, **kwargs):
    f = impl(*args, **kwargs)
    shape_and_dtype = spec(*args, **kwargs)
    g = grid(*args, **kwargs)
    if multiple_results:
      for fi, (si, di), gi in zip(f, shape_and_dtype, g):
        _set_shape_and_dtype(fi, si, di)
        _set_grid(fi, gi)
    else:
      _set_shape_and_dtype(f, *shape_and_dtype)
      _set_grid(f, g)
    return f

  @prim.def_abstract_eval
  def _abstract_eval(*args, **kwargs):
    shape_and_dtype = spec(*args, **kwargs)
    if multiple_results:
      return tuple(
        GeneralArray(s) if d is None else core.ShapedArray(s, dtype=d)
        for s, d in shape_and_dtype
      )
    else:
      s, d = shape_and_dtype
      return (
        GeneralArray(s)
        if shape_and_dtype[1] is None else core.ShapedArray(s, dtype=d)
      )


def _assert_same_input_signature(*fs):
  arg_specs = tuple(SpecTree.from_args(f) for f in fs)
  for d in arg_specs[1:]:
    if d != arg_specs[0]:
      raise ValueError("All functions must have the same input signature.")


def _assert_same_output_signature(*fs):
  arg_specs = tuple(SpecTree.from_ret(f) for f in fs)
  for d in arg_specs[1:]:
    if d != arg_specs[0]:
      raise ValueError("All functions must have the same output signature.")


def _assert_composable(f, *gs):
  f_i_spec = SpecTree.from_args(f)
  gs_o_spec = tuple(SpecTree.from_ret(g) for g in gs)
  if f_i_spec != gs_o_spec:
    raise ValueError(
      f"Cannot compose functions with incompatible shapes: "
      f"f requires: {SpecTree.from_args(f)}, "
      f"while gs provides {tuple(SpecTree.from_ret(g) for g in gs)}"
    )


@operator
@lu.wrap_init
def partial_impl(f, *, args, argnums):
  assert len(argnums) == len(args)

  @function
  @lu.wrap_init
  def partial_f(*rest_args):
    all_args = [None] * f.num_args
    a, b = 0, 0
    for i in range(f.num_args):
      if a < len(argnums) and i == argnums[a]:
        all_args[i] = args[a]
        a += 1
      else:
        all_args[i] = rest_args[b]
        b += 1
    return f(*all_args)

  return partial_f


# permute_args


def permute_args(f, *, perm):
  return permute_args_p.bind(f, perm=tuple(perm))


@operator
@lu.wrap_init
def permute_args_impl(f, *, perm):
  return function(t.permute_args(f, perm))


def permute_args_spec(f, *, perm):
  return (f.ret, *(f.arg[i] for i in perm)), None


def permute_args_grid(f, *, perm):
  return f.grid.take(perm)


permute_args_p = core.Primitive("permute_args")
def_operator(
  permute_args_p,
  permute_args_impl,
  permute_args_spec,
  permute_args_grid,
)


def _permute_args_transpose_rule(t, f, *, perm):
  assert ad.is_undefined_primal(f)
  inv_perm = np.argsort(perm)
  return (permute_args_p.bind(t, perm=tuple(inv_perm.tolist())),)


jax.interpreters.ad.deflinear2(permute_args_p, _permute_args_transpose_rule)
jax.interpreters.ad.primitive_transposes[permute_args_p
                                        ] = _permute_args_transpose_rule

# split


def split(f):
  return split_p.bind(f)


@operator
@lu.wrap_init
def split_impl(f):
  return tuple(function(t.return_at(f, i)) for i in range(len(f.ret.spec)))


def split_spec(f):
  assert isinstance(f.ret.spec, tuple)
  return tuple(((r, *f.arg), None) for r in map(Ret, f.ret.spec))


def split_grid(f):
  return tuple(f.grid for _ in f.ret.spec)


split_p = core.Primitive("split")
def_operator(
  split_p,
  split_impl,
  split_spec,
  split_grid,
  multiple_results=True,
)


def _split_transpose_rule(t, f):
  return (zip_functions(*t, share_inputs=True),)


jax.interpreters.ad.deflinear2(split_p, _split_transpose_rule)
jax.interpreters.ad.primitive_transposes[split_p] = _split_transpose_rule

# zip_functions


def zip_functions(*fs, share_inputs=False):
  """Zip functions.
  This requires all `*fs` to have the same input signature.
  Returns a function that returns a tuple of the outputs of the functions.

  Equivalent to:

  .. code-block:: python

    def zip_functions(*fs):
      def zipped_fs(*args):
        return tuple(f(*args) for f in fs)
      return zipped_fs

  Args:
    *fs: functions to be zipped.
  Returns:
    A function that returns a tuple of the outputs of the functions.
  """
  return compose(
    HashableFunction(lambda *args: args, None),
    *fs,
    share_inputs=share_inputs,
  )


# broadcast_to


def broadcast_arg(f, target, *, argnums):
  argnums = (argnums,) if isinstance(argnums, int) else argnums
  return broadcast_arg_p.bind(f, target, argnums=argnums)


def broadcast_arg_impl(f, target, *, argnums):
  return function(t.broadcast_arg(f, target.shape, argnums))


def broadcast_arg_spec(f, target, *, argnums):
  assert f.ret == target.shape[0]
  return target.shape, None


def broadcast_arg_grid(f, target, *, argnums):
  c_argnums = complement_argnums(argnums, target.num_args)
  if hasattr(target, "grid"):
    grid = f.grid + target.grid.take(c_argnums)
  else:
    grid = f.grid + Grid.none(len(c_argnums))
  perm = np.argsort(argnums + c_argnums)
  return grid.take(perm)


broadcast_arg_p = core.Primitive("broadcast_arg")

def_operator(
  broadcast_arg_p,
  broadcast_arg_impl,
  broadcast_arg_spec,
  broadcast_arg_grid,
)


def _broadcast_arg_jvp_rule(primals, tangents, *, argnums):
  f, target = primals
  f_dot, target_dot = tangents
  primals_out = broadcast_arg_p.bind(f, target, argnums=argnums)
  if not isinstance(f_dot, ad_util.Zero):
    tangent_out = broadcast_arg_p.bind(f_dot, target, argnums=argnums)
  else:
    tangent_out = ad_util.Zero.from_value(target_dot)
  return primals_out, tangent_out


def _broadcast_arg_transpose_rule(t, f, target, *, argnums):
  c_argnums = complement_argnums(argnums, target.num_args)
  return (integrate(t, argnums=c_argnums), None)


jax.interpreters.ad.primitive_jvps[broadcast_arg_p] = _broadcast_arg_jvp_rule
jax.interpreters.ad.primitive_transposes[broadcast_arg_p
                                        ] = _broadcast_arg_transpose_rule

# linear_transpose


def linear_transpose(f, *, argnums=0, unpack=False):
  """Returns the transpose of a linear function."""
  return linear_transpose_p.bind(
    f,
    argnums=argnums,
    unpack=unpack,
  )


@operator
@lu.wrap_init
def linear_transpose_impl(f, *, argnums, unpack):
  assert isinstance(f, function)
  argnums_tuple = (argnums,) if isinstance(argnums, int) else argnums
  bound_argnums = complement_argnums(argnums, f.num_args)
  dummy_spec = tuple(map(lambda idx: f.arg_spec[idx], argnums_tuple))
  dummy_primals = dummy_array(dummy_spec)

  @function
  @lu.wrap_init
  def transposed_f(*args):
    if not unpack:
      bound_args, cotangent = args[:-1], args[-1]
    else:
      bound_args, cotangent = args[:-len(f.ret_spec)], args[-len(f.ret_spec):]
    partial_f = partial_impl(f, args=bound_args, argnums=bound_argnums)
    ret = jax.linear_transpose(partial_f.f.call_wrapped, *dummy_primals)(
      cotangent
    )
    return ret[0] if isinstance(argnums, int) else tuple(ret)

  return transposed_f


def linear_transpose_spec(f, *, argnums, unpack):
  c_argnums = complement_argnums(argnums, f.num_args)
  if isinstance(argnums, int):
    ret = f.arg[argnums]
  else:
    ret = Ret(tuple(f.arg[i].spec for i in argnums))
  if unpack:
    assert isinstance(f.ret.spec, tuple)
    return (
      ret, *map(f.arg.__getitem__, c_argnums), *map(Arg, f.ret.spec)
    ), None
  else:
    return (ret, *map(f.arg.__getitem__, c_argnums), f.ret), None


def linear_transpose_grid(f, *, argnums, unpack):
  c_argnums = complement_argnums(argnums, f.num_args)
  if unpack:
    return f.grid.take(c_argnums) + Grid.none(len(f.ret.spec))
  else:
    return f.grid.take(c_argnums) + Grid.none(1)


linear_transpose_p = core.Primitive("linear_transpose")
def_operator(
  linear_transpose_p,
  linear_transpose_impl,
  linear_transpose_spec,
  linear_transpose_grid,
)


def _linear_transpose_transpose_rule(t, f, *, argnums, unpack):
  assert ad.is_undefined_primal(f)
  o_spec = SpecTree.from_ret(f)
  argnums_tuple = (argnums,) if isinstance(argnums, int) else argnums
  rest_index = complement_argnums(argnums, num_args(f))
  # if the primal function unpacks its output into the args
  # of the cotangent function
  if not unpack:
    argnums_t = num_args(t) - 1
  else:
    n = num_args(t)
    argnums_t = tuple(range(n - len(o_spec), n))
  unpack_t = (not isinstance(argnums, int))
  ret = linear_transpose_p.bind(
    t,
    argnums=argnums_t,
    unpack=unpack_t,  # unpack if argnums is a tuple
  )
  perm = np.argsort(rest_index + argnums_tuple)
  ret = permute_args_p.bind(ret, perm=tuple(perm))
  return (ret,)


jax.interpreters.ad.deflinear2(
  linear_transpose_p, _linear_transpose_transpose_rule
)
jax.interpreters.ad.primitive_transposes[linear_transpose_p
                                        ] = _linear_transpose_transpose_rule

# integrate


def integrate(f, *, argnums=0):
  """Integrate function f over all input arguments.
  WARNING, the forward pass of this function is not implemented.
  This is only useful for computing the functional gradient of
  integral type functionals.

  Args:
    f: Function to integrate.
  Returns:
    Scalar representing the integral of f.
  """
  return integrate_p.bind(f, argnums=argnums)


@operator
@lu.wrap_init
def integrate_impl(f, *, argnums):
  if num_args(f) == 0:
    return f()
  argnums = (argnums,) if isinstance(argnums, int) else argnums
  c_argnums = complement_argnums(argnums, f.num_args)
  grid = f.grid.take(argnums)

  @lu.wrap_init
  def partially_integrated_f(*args):
    nargs = num_args(f)
    vf = f
    for argidx in argnums:
      in_axes = tuple(0 if i == argidx else None for i in range(nargs))
      vf = jax.vmap(vf, in_axes=in_axes, out_axes=0)
    idx = np.argsort(c_argnums + argnums)
    concat_args = (args + grid.nodes)
    full_args = [concat_args[i] for i in idx]
    out = vf(*full_args)
    for w in grid.weights:
      out = w @ out
    return out

  if any(n is None for n in grid.nodes):
    if len(c_argnums) == 0:
      logging.warning(
        "grids_and_weights is None, this will create a dummy integration "
        "that always returns 0. The gradient of this dummy integration would "
        "still work."
      )
      return tree_map(lambda s: jnp.zeros(s.shape, s.dtype), f.ret.spec)
    else:
      raise ValueError("grids_and_weights needs to be specified.")
  else:
    if len(c_argnums) == 0:
      return partially_integrated_f.call_wrapped()
    else:
      return function(partially_integrated_f)


def integrate_spec(f, *, argnums):
  c_argnums = complement_argnums(argnums, f.num_args)
  rest_args = tuple(map(f.arg.__getitem__, c_argnums))
  if len(rest_args) == 0:
    return f.ret.spec.shape, f.ret.spec.dtype
  else:
    return (f.ret, *rest_args), None


def integrate_grid(f, *, argnums):
  c_argnums = complement_argnums(argnums, f.num_args)
  return f.grid.take(c_argnums)


integrate_p = core.Primitive("integrate")
def_operator(
  integrate_p,
  integrate_impl,
  integrate_spec,
  integrate_grid,
)


def _integrate_jvp_rule(primals, tangents, *, argnums):
  f = primals[0]
  f_dot = tangents[0]
  primals_out = integrate_p.bind(f, argnums=argnums)
  if not isinstance(f_dot, ad_util.Zero):
    tangents_out = integrate_p.bind(f_dot, argnums=argnums)
  else:
    tangents_out = ad_util.Zero.from_value(primals_out)
  return primals_out, tangents_out


def _integrate_transpose_rule(t, f, *, argnums):
  assert ad.is_undefined_primal(f)
  if is_function(t):
    cargnums = complement_argnums(argnums, num_args(f))
    return (broadcast_arg(t, f.aval, argnums=cargnums),)
  else:

    @with_spec(SpecTree.from_args(f), SpecTree.from_ret(f))
    def return_t(*args):
      return t

    return (return_t,)


jax.interpreters.ad.primitive_jvps[integrate_p] = _integrate_jvp_rule
jax.interpreters.ad.primitive_transposes[integrate_p
                                        ] = _integrate_transpose_rule

# compose


def compose(f, *gs, share_inputs=False):
  """Compose function f with all the gs.

  Equivalent to:

  .. code-block:: python

    def compose(f, *gs):
      def composed(*args):
        return f(
          *(g(*arg) for arg, g in zip(args, gs))
        )
      return composed

  Args:
    f: function to compose with.
    *gs: functions to compose f with.
  Returns:
    A function that applies f to the outputs of gs.
  """
  if not isinstance(f, function) and not isinstance(f, jax.core.Tracer):
    if not callable(f):
      raise ValueError(f"f must be a function. Got: {f}")
    arg_spec = tuple(SpecTree.from_ret(g) for g in gs)
    try:
      ret_spec = SpecTree.from_value(jax.eval_shape(f, *arg_spec))
    except Exception as e:
      raise ValueError(
        "Cannot compose functions because applying "
        f"{f} on {', '.join(map(str, arg_spec))} fails."
      ) from e
    f = with_spec(arg_spec=arg_spec, ret_spec=ret_spec)(f)
  return compose_p.bind(f, *gs, share_inputs=share_inputs)


def compose_impl(f, *gs, share_inputs):
  return function(t.compose(f, gs, tuple(g.num_args for g in gs), share_inputs))


def compose_spec(f, *gs, share_inputs):
  _assert_composable(f, *gs)
  if share_inputs:
    _assert_same_input_signature(*gs)
    return (f.ret, *gs[0].arg), None
  return (f.ret, *(a for g in gs for a in g.arg)), None


def compose_grid(f, *gs, share_inputs):
  if share_inputs:
    return gs[0].grid
  else:
    return sum(g.grid for g in gs)


compose_p = core.Primitive("compose")
def_operator(
  compose_p,
  compose_impl,
  compose_spec,
  compose_grid,
)


def _compose_jvp_rule(primals, tangents, *, share_inputs):
  f, gs = primals[0], primals[1:]
  f_dot, gs_dot = tangents[0], tangents[1:]
  primals_out = compose_p.bind(f, *gs, share_inputs=share_inputs)
  tangents_out = []
  # contribution from f_dot
  if not isinstance(f_dot, ad_util.Zero):
    tangents_out.append(compose_p.bind(f_dot, *gs, share_inputs=share_inputs))

  # contribution from gs_dot
  nonzero_gs_dot = []
  nonzero_indice = []
  argnums = []
  argidx = 0
  for i, g_dot in enumerate(gs_dot):
    if not isinstance(g_dot, ad_util.Zero):
      nonzero_gs_dot.append(g_dot)
      nonzero_indice.append(i)
      argnums.extend(list(range(argidx, argidx + num_args(g_dot))))
    argidx += num_args(g_dot)
  argnums = tuple(argnums)

  gs = zip_functions(*gs, share_inputs=share_inputs)
  gs_dot = zip_functions(*nonzero_gs_dot, share_inputs=share_inputs)
  if not share_inputs:
    gs_dot = broadcast_arg(gs_dot, gs, argnums=argnums)

  tangents_out.append(
    compose_p.bind(
      linearize_p.bind(f, argnums=tuple(nonzero_indice)),
      *split_p.bind(gs),
      *split_p.bind(gs_dot),
      share_inputs=True,
    )
  )

  # one more step to sum contributions from f_dot and gs_dot
  return primals_out, add(*tangents_out)


def _compose_transpose_rule(cotangent, f, *gs, share_inputs):

  if isinstance(cotangent, ad_util.Zero):
    return (None, *(ad_util.Zero.from_value(a) for a in gs))

  if ad.is_undefined_primal(f):
    # when gs are fixed, the compose rule is linear to f
    # because compose(f1, *gs) + compose(f2, *gs) = compose(f1 + f2, *gs)
    # cotangent: A -> C, gs: A -> B.
    # However, this requires us to know the inverse of gs,
    # so that we construct inverse g: B -> A with cotangent: A -> C into B -> C
    # We currently just disable the transpose rule wrt f.
    # TODO: maybe when gs are invertible, we could enable it.
    raise NotImplementedError(
      "transpose of compose(f, g1, g2, ...) wrt f is not defined."
    )
  else:
    # when f is linear, the compose rule is linear to each gs
    # cotangent: A -> C, f: B -> C  => A -> B
    # we need to apply linear_transpose on function f elementwisely.
    # Then, we can compose(linear_transposed_f, cotangent),
    # which generates a function of A -> B.
    # This could fail if f is not linear.
    narg = num_args(f)
    ret = [None] * (1 + narg)
    argnums, rest_args = [], []
    for i, g in enumerate(gs):
      if ad.is_undefined_primal(g):
        argnums.append(i)
      else:
        rest_args.append(g)
        ret[1 + i] = None
    argnums = tuple(argnums)
    if share_inputs or len(gs) == 1:
      wrt_tangent = compose_p.bind(
        linear_transpose(f, argnums=argnums),
        *rest_args,
        cotangent,
        share_inputs=share_inputs,
      )
    else:
      # TODO: implementing this requires a general method of
      # performing intergration, keep for future.
      # we first broadcast the args of the rest_args to the same as
      # the cotangent, and then compose the transposed f in the same way
      # as in the above share_inputs=True branch.
      # finally, we integrate over the args of the rest_args.
      raise NotImplementedError
    for i, tgt in zip(argnums, split_p.bind(wrt_tangent)):
      ret[1 + i] = tgt
    return tuple(ret)


jax.interpreters.ad.primitive_jvps[compose_p] = _compose_jvp_rule
jax.interpreters.ad.primitive_transposes[compose_p] = _compose_transpose_rule

# nabla


def nabla(f, *, argnums=0, method=jax.jacfwd):
  """nabla operator obtains the gradient function.
  In the simplest case, we consider single variable single output function.
  f: R -> R, (nabla f): R -> R.
  While we can generalize to multiple variables and multiple outputs.
  for f: R^n -> R^m, (nabla f): R^n -> R^(m x n).

  Equivalent to:

  .. code-block:: python

    def nabla(f, argnums=0, has_aux=False):
      def nabla_f(*args):
        jax.jacfwd(f, argnums, has_aux=has_aux)(*args)
      return nabla_f

  Args:
    f: function to be differentiated.
    argnums: the argument index to be differentiated.
    has_aux: whether the function has auxilary outputs.
  Returns:
    (nabla f): the gradient function.
  """
  assert method in (
    jax.jacfwd, jax.jacrev
  ), "method need to be jax.jacfwd or jax.jacrev"
  return nabla_p.bind(f, argnums=argnums, method=method)


@operator
@lu.wrap_init
def nabla_impl(f, *, argnums, method):
  return function(lu.wrap_init(method(f, argnums)))


def nabla_spec(f, *, argnums, method):
  ispec = f.arg[argnums].spec if isinstance(argnums, int) else tuple(
    f.arg[i].spec for i in argnums
  )
  ospec = f.ret.spec
  jspec = tree_map(
    lambda osp: tree_map(
      lambda isp: Spec(
        osp.shape + isp.shape, osp.dtype if method == jax.jacfwd else isp.dtype
      ), ispec
    ), ospec
  )
  return (Ret(jspec), *f.arg), None


def nabla_grid(f, *, argnums, method):
  return f.grid


nabla_p = core.Primitive("nabla")
def_operator(
  nabla_p,
  nabla_impl,
  nabla_spec,
  nabla_grid,
)


def _nabla_transpose_rule(t, f, *, argnums, method):
  cotangent_spec = SpecTree.from_ret(t)
  assert ad.is_undefined_primal(f)
  jac_spec = jacobian_spec(f.aval, argnums)
  if (cotangent_spec != jac_spec):
    raise ValueError(
      f"The cotangent is expected to be {jac_spec}, got {cotangent_spec}."
    )
  o_spec = SpecTree.from_ret(f)
  nabla_t = nabla(t, argnums=argnums, method=method)

  @with_spec((SpecTree.from_ret(nabla_t),), SpecTree.from_ret(f))
  def _contract(arg, /):
    """The transpose of nabla is simply -nabla when the operand is
    a scalar->scalar function. However, here we generalized to the
    case where both input/output of the function can be a pytree of
    arrays.
    """

    # negative nabla
    # arg = tree_map(lambda y: -y, arg)

    def neg_sum_diag(spec_prefix, xf):
      xf = tree_leaves(xf)
      ll = int(np.sqrt(len(xf)))
      out = []
      for i in range(ll):
        x = xf[i + i * ll]
        assert x.shape[:len(spec_prefix.shape)] == spec_prefix.shape
        l = int(np.sqrt(np.prod(x.shape[len(spec_prefix.shape):])))
        out.append(
          jnp.trace(x.reshape(spec_prefix.shape + (l, l)), axis1=-2, axis2=-1)
        )
      return -sum(out)

    out = tree_map(neg_sum_diag, o_spec, arg)
    return out

  ret = compose(_contract, nabla_t)
  return (ret,)


def _nabla_jvp_rule(primals, tangents, *, argnums, method):
  primal_out = nabla(primals[0], argnums=argnums, method=method)
  tangent_out = nabla(tangents[0], argnums=argnums, method=method)
  return primal_out, tangent_out


jax.interpreters.ad.primitive_jvps[nabla_p] = _nabla_jvp_rule
jax.interpreters.ad.primitive_transposes[nabla_p] = _nabla_transpose_rule

# linearize


def linearize(f, *, argnums=0):
  """Linearize is a bit different from the jax.linearize call
  The reason is that jax.linearize requires both function and the
  primal values. As an operator, the linearize function here takes
  as input a function f. And returns a function lf(primal, tangent).

  Equivalent to:

  .. code-block:: python

    def linearize(f):
      def linearized_f(primal, tangent):
        return jax.jvp(f, primal, tangent)
      return linearized_f

  Args:
    f: function to be differentiated.
    argnums: the argument index to be differentiated.
      default to None, which computes linearization wrt all arguments.
  Returns:
    (linear f): the linearized function that takes primal and tangent.
  """
  argnums = (argnums,) if isinstance(argnums, int) else argnums
  return linearize_p.bind(f, argnums=argnums)


@operator
@lu.wrap_init
def linearize_impl(f, argnums):
  assert isinstance(f, function)

  @function
  @lu.wrap_init
  def linearized_f(*args):
    assert len(args) == f.num_args + len(argnums)
    primals, tangents = args[:f.num_args], args[f.num_args:]
    bound_argnums = complement_argnums(argnums, f.num_args)
    unbound_args = tuple(map(primals.__getitem__, argnums))
    bound_args = tuple(map(primals.__getitem__, bound_argnums))
    partial_f = partial_impl(f, args=bound_args, argnums=bound_argnums)
    return jax.linearize(partial_f.f.call_wrapped, *unbound_args)[1](*tangents)

  return linearized_f


def linearize_spec(f, argnums):
  return (f.ret, *f.arg, *map(f.arg.__getitem__, argnums)), None


def linearize_grid(f, argnums):
  return f.grid + Grid.none(len(argnums))


linearize_p = core.Primitive("linearize")
def_operator(
  linearize_p,
  linearize_impl,
  linearize_spec,
  linearize_grid,
)


def _linearize_transpose_rule(t, f, *, argnums):
  # t: (primal, tangent) -> (primal_out, tangent_out)
  # f: (primal -> out)
  # seems we need to integrate over one of the inputs.
  # not very sure as this will go to infinity.
  raise NotImplementedError


def _linearize_jvp_rule(primals, tangents, *, argnums):
  # this is supposed to be an linear operator
  # because it consists of applying a nabla operator on the input function
  # and then product with an identity function.
  f_dot = tangents[0]
  return linearize_p.bind(f_dot, argnums=argnums)


jax.interpreters.ad.primitive_jvps[linearize_p] = _linearize_jvp_rule
jax.interpreters.ad.primitive_transposes[linearize_p
                                        ] = _linearize_transpose_rule


def constant(*, value, arg_shape):
  return constant_p.bind(value=value, arg_shape=arg_shape)


@operator
@lu.wrap_init
def constant_impl(*, value, arg_shape):

  @function
  @lu.wrap_init
  def const_value(*args):
    return value

  return const_value


def constant_spec(*, value, arg_shape):
  return (ret_like(value), *arg_shape), None


def constant_grid(*, value, arg_shape):
  return Grid.none(len(arg_shape))


constant_p = core.Primitive("constant")
def_operator(
  constant_p,
  constant_impl,
  constant_spec,
  constant_grid,
)


def partial(f, *, args, argnums=0):
  argnums = (argnums,) if isinstance(argnums, int) else argnums
  return partial_p.bind(f, args=args, argnums=argnums)


def partial_spec(f, *, args, argnums):
  assert len(argnums) == len(args)
  c_argnums = complement_argnums(argnums, num_args(f))
  return (f.ret, *(f.arg[i] for i in c_argnums)), None


def partial_grid(f, *, args, argnums):
  c_argnums = complement_argnums(argnums, num_args(f))
  return f.grid.take(c_argnums)


partial_p = core.Primitive("partial")
def_operator(
  partial_p,
  partial_impl,
  partial_spec,
  partial_grid,
)


def _partial_transpose_rule(t, f, *, args, argnums):
  argnums = (argnums,) if isinstance(argnums, int) else argnums
  cargnums = complement_argnums(argnums, num_args(f))
  assert ad.is_undefined_primal(f)
  wrt_f = broadcast_arg(t, f, cargnums)
  return (wrt_f,)


def _partial_jvp_rule(primals, tangents, *, args, argnums):
  f = primals[0]
  f_dot = tangents[0]
  jvps = []
  if not isinstance(f_dot, ad_util.Zero):
    jvp1 = partial(f_dot, args=args, argnums=argnums)
    jvps.append(jvp1)
  return partial(f, args=args, argnums=argnums), sum(jvps)


ad.primitive_jvps[partial_p] = _partial_jvp_rule
ad.primitive_transposes[partial_p] = _partial_transpose_rule


def braket(lfun, rfun, *, real=False):
  """Braket operator.

  Equivalent to:

  .. code-block:: python
    intergrate(numpy.conj(lfun) * rfun)

  Args:
    lfun: left function.
    rfun: right function.

  Returns:
    A scalar that represents the integral of the product of the
    conjugate of lfun and rfun.
  """
  return braket_p.bind(lfun, rfun, real=real)


def braket_impl(lfun, rfun, *, real):
  argnums = tuple(range(lfun.num_args))
  if real:
    prod = lambda l, r: jnp.real(jnp.conj(l) * r)
  else:
    prod = lambda l, r: jnp.conj(l) * r
  integrand = compose(prod, lfun, rfun, share_inputs=True)
  return integrate_impl(integrand, argnums=argnums)


def braket_spec(lfun, rfun, *, real):
  assert lfun.shape == rfun.shape
  shape = lfun.ret.spec.shape
  if real:
    dtype = jnp.zeros((), dtype=lfun.ret.spec.dtype).astype(float).dtype
  else:
    dtype = lfun.ret.spec.dtype
  return shape, dtype


def bracket_grid(lfun, rfun, *, real):
  return lfun.grid


braket_p = core.Primitive("braket")
def_operator(
  braket_p,
  braket_impl,
  braket_spec,
  bracket_grid,
)


def _braket_jvp_rule(primals, tangents, *, real):
  lfun, rfun = primals
  lfun_dot, rfun_dot = tangents
  primals_out = braket_p.bind(lfun, rfun, real=real)
  tangents_out = []
  if not isinstance(lfun_dot, ad_util.Zero):
    tangents_out.append(braket_p.bind(lfun_dot, rfun, real=real))
  if not isinstance(rfun_dot, ad_util.Zero):
    tangents_out.append(braket_p.bind(lfun, rfun_dot, real=real))
  return primals_out, sum(tangents_out)


def _braket_transpose_rule(t, lfun, rfun, *, real):
  assert not (ad.is_undefined_primal(lfun) and ad.is_undefined_primal(rfun))
  if ad.is_undefined_primal(lfun):
    t = constant(value=t, arg_shape=rfun.shape[1:])
    return (t * rfun, None)
  elif ad.is_undefined_primal(rfun):
    t = constant(value=t, arg_shape=lfun.shape[1:])
    return (None, t * lfun)


jax.interpreters.ad.primitive_jvps[braket_p] = _braket_jvp_rule
jax.interpreters.ad.primitive_transposes[braket_p] = _braket_transpose_rule


def invoke(f, *args):
  if len(args) != num_args(f):
    raise ValueError(
      f"Function {f} expects {num_args(f)} arguments, got {len(args)}."
    )
  return integrate(partial(f, args=args, argnums=tuple(range(len(args)))))


def binary_op(op, f1, f2, **params):
  op = functools.partial(op, **params)
  if not is_function(f1) and not is_function(f2):
    raise ValueError(
      "This will not happen because binary op is bind to at least one function."
    )
  elif is_function(f1) and not is_function(f2):
    if isinstance(f2, (tuple, list)):
      return compose(
        HashableFunction(lambda x: tree_map(op, x, f2), (op, f2)),
        f1,
      )
    else:
      return compose(
        HashableFunction(lambda x: tree_map(lambda y: op(y, f2), x), (op, f2)),
        f1,
      )
  elif is_function(f2) and not is_function(f1):
    if isinstance(f1, (tuple, list)):
      return compose(
        HashableFunction(lambda x: tree_map(op, f1, x), (op, f1)), f2
      )
    else:
      return compose(
        HashableFunction(lambda x: tree_map(lambda y: op(f1, y), x), (op, f1)),
        f2,
      )
  else:
    ret_f1, ret_f2 = SpecTree.from_ret(f1), SpecTree.from_ret(f2)
    is_pytree = lambda spec: isinstance(spec, (tuple, list))
    if (
      is_pytree(ret_f1) and is_pytree(ret_f2) and
      tree_structure(ret_f1) != tree_structure(ret_f2)
    ):
      raise ValueError(
        f"Cannot apply {op} on functions with different return structure. "
        f"Got {ret_f1} and {ret_f2}"
      )
    if SpecTree.from_args(f1) != SpecTree.from_args(f2):
      raise ValueError(
        f"Cannot apply {op} on functions with different arg specs. "
        f"Got {SpecTree.from_args(f1)} and {SpecTree.from_args(f2)}"
      )
    if is_pytree(ret_f1) and not is_pytree(ret_f2):
      return compose(
        HashableFunction(
          lambda tree, y: tree_map(lambda x: op(x, y), tree), (op,)
        ),
        f1,
        f2,
        share_inputs=True,
      )
    elif is_pytree(ret_f2) and not is_pytree(ret_f1):
      return compose(
        lambda x, tree: tree_map(lambda y: op(x, y), tree),
        f1,
        f2,
        share_inputs=True,
      )
    else:
      return compose(lambda x, y: tree_map(op, x, y), f1, f2, share_inputs=True)


def unary_op(op, f, **params):
  op = functools.partial(op, **params)
  if not is_function(f):
    raise ValueError(
      "This will not happen because unary op is bind to a function."
    )
  return compose(lambda x: tree_map(op, x), f, share_inputs=True)


funcs = {
  "numpy":
    {
      "negative": unary_op,
      "exp": unary_op,
      "log": unary_op,
      "sin": unary_op,
      "cos": unary_op,
      "tan": unary_op,
      "arcsin": unary_op,
      "arccos": unary_op,
      "arctan": unary_op,
      "sinh": unary_op,
      "cosh": unary_op,
      "tanh": unary_op,
      "arcsinh": unary_op,
      "arccosh": unary_op,
      "arctanh": unary_op,
      "square": unary_op,
      "sqrt": unary_op,
      "abs": unary_op,
      "sum": unary_op,
      "prod": unary_op,
      "mean": unary_op,
      "max": unary_op,
      "min": unary_op,
      "argmax": unary_op,
      "argmin": unary_op,
      "sign": unary_op,
      "add": binary_op,
      "subtract": binary_op,
      "multiply": binary_op,
      "divide": binary_op,
      "power": binary_op,
    },
  "nn": {
    "sigmoid": unary_op,
    "relu": unary_op,
  },  # noqa
  "lax": {
    "erf": unary_op,
  },  # noqa
}

numpy = SimpleNamespace()
lax = SimpleNamespace()
nn = SimpleNamespace()

for namespace, fns in funcs.items():
  ns = globals()[namespace]
  for name, nary_op in fns.items():
    jax_fn = getattr(getattr(jax, namespace), name)
    setattr(ns, name, functools.partial(nary_op, jax_fn))


def add(*fs):
  """Add functions together.

  Equivalent to:

  .. code-block:: python

    def add(*fs):
      def summed(*args):
        return sum([f(*args) for f in fs])
      return summed

  Args:
    x, y: functions to add together.
  Returns:
    A function that sums the outputs of the functions.
  """
  if len(fs) > 1:

    def add_values(*args):
      return tree_map(lambda *xs: sum(xs), *args)

    return compose(add_values, *fs, share_inputs=True)
  elif len(fs) == 1:
    return fs[0]


ad_util.raw_jaxval_adders[types.FunctionType] = add
ad_util.raw_jaxval_adders[function] = add

array_operators = {
  "neg": lambda x: numpy.negative(x),  # noqa
  "add": lambda x, y: numpy.add(x, y),
  "radd": lambda y, x: numpy.add(x, y),
  "sub": lambda x, y: numpy.subtract(x, y),
  "rsub": lambda y, x: numpy.subtract(x, y),
  "truediv": lambda x, y: numpy.divide(x, y),
  "rtruediv": lambda y, x: numpy.divide(x, y),
  "mul": lambda x, y: numpy.multiply(x, y),
  "rmul": lambda y, x: numpy.multiply(x, y),
  "pow": lambda x, y: numpy.power(x, y),
  "rpow": lambda y, x: numpy.power(x, y),
}


def call(obj, *args, **params):
  args = tuple(
    jnp.asarray(a) if isinstance(a, (float, int)) else a for a in args
  )
  # check all inputs are correct
  arg_spec = tuple(
    SpecTree.from_ret(a) if is_function(a) else SpecTree.from_value(a)
    for a in args
  )
  if arg_spec != obj.arg_spec:
    raise ValueError(
      f"Spec mismatch: function {obj.f} is marked as "
      f"requiring {obj.arg_spec} as input, got {arg_spec}."
    )
  # separate function and non-function arguments
  non_function_args = []
  non_function_index = []
  function_args = []
  for i, arg in enumerate(args):
    if is_function(arg):
      function_args.append(arg)
    else:
      non_function_args.append(arg)
      non_function_index.append(i)

  # if all are non functions, call invoke
  if len(function_args) == 0:
    if isinstance(obj, function):
      # if this is a function instance, directly invoke it.
      ret = obj._call(*args)  # noqa
    else:
      # if this is abstract, use the invoke operator instead.
      ret = invoke(obj, *args)
    if SpecTree.from_value(ret) != obj.ret_spec:
      raise ValueError(
        f"Spec mismatch: function {obj.f} is marked as "
        f"returning {obj.ret_spec}, got {SpecTree.from_value(ret)}."
      )
    return ret
  # if there are both function and non-function arguments
  # first bind the non-function arguments
  # then compose with the function arguments
  elif len(non_function_args) != 0:
    f = partial(obj, args=non_function_args, argnums=tuple(non_function_index))
  # if all are function arguments, just compose
  else:
    f = obj
  return compose(f, *function_args, **params)


def _set_function_operators(cls, operators):
  cls.__call__ = call
  for op_name, op in operators.items():
    setattr(cls, f"__{op_name}__", op)


def _set_general_array_operators(cls, operators):
  cls.__call__ = call
  for op_name, op in operators.items():
    setattr(cls, f"_{op_name}", staticmethod(op))


def _set_tracer_call():

  def call(self, *args):
    if isinstance(self.aval, GeneralArray):
      return invoke(self, *args)
    else:
      raise ValueError("Calling an array")

  jax.core.Tracer.__call__ = call


_set_function_operators(function, array_operators)
_set_general_array_operators(GeneralArray, array_operators)
_set_tracer_call()



#!/usr/bin/env python3
# Copyright 2023 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
  import jax
  import optax

  def scale_by_learning_rate(
    learning_rate,
    *,
    flip_sign: bool = True,
  ):
    m = -1 if flip_sign else 1
    # avoid calling tracer
    if callable(learning_rate
               ) and not isinstance(learning_rate, jax.core.Tracer):
      return optax._src.transform.scale_by_schedule(
        lambda count: m * learning_rate(count)
      )
    return optax._src.transform.scale(m * learning_rate)

  optax._src.alias._scale_by_learning_rate = scale_by_learning_rate
except ImportError:
  print("optax not install, skip patching")

from . import operators  # noqa
from .general_array import SpecTree  # noqa
from .general_array import (
  Arg, Grid, Ret, Spec, dummy_array, dummy_input, dummy_output, function,
  is_function, num_args, random_input, with_spec, zeros_like
)

__all__ = [
  "Spec",
  "SpecTree",
  "Arg",
  "Ret",
  "Grid",
  "function",
  "with_spec",
  "num_args",
  "zeros_like",
  "is_function",
  "dummy_input",
  "dummy_output",
  "random_input",
  "dummy_array",
  "operators",
]

__version__ = "0.0.8"  # noqa



# Copyright 2023 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""lu transformations"""

import numpy as np
import jax.extend.linear_util as lu


@lu.transformation2
def permute_args(f, perm, *args):
  inv_perm = tuple(np.argsort(perm))
  return f(*(args[i] for i in inv_perm))


@lu.transformation2
def return_at(f, i, *args):
  return f(*args)[i]


@lu.transformation2
def compose(f, gs, num_args, share_inputs, *args):
  i = 0
  if share_inputs:
    fargs = tuple(g(*args) for g in gs)
  else:
    fargs = []
    for g, na in zip(gs, num_args):
      fargs.append(g(*args[i:i + na]))
      i += na
    fargs = tuple(fargs)
  return f(*fargs)


@lu.transformation2
def broadcast_arg(f, shape, argnums, *args):
  args_shape = shape[1:]
  assert len(args_shape) == len(args)
  # assert tuple(Arg(Spec(a.shape, a.dtype)) for a in args) == args_shape
  return f(*(args[i] for i in argnums))



#!/usr/bin/env python3
# Copyright 2023 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jaxtyping import Float32, Array
from typing import Tuple
from absl.testing import absltest, parameterized
from autofd.general_array import (
  Arg,
  Ret,
  Spec,
  function,
  jacobian_spec,
  SpecTree,
  dummy_input,
)


@function
def function0(x: Float32[Array, "3 5"]) -> Float32[Array, ""]:
  return x.sum()


@function
def function1(x: Float32[Array, "3 5"]) -> Float32[Array, "3"]:
  return x.sum(1)


@function
def function2(
  x: Float32[Array, "3 5"]
) -> Tuple[Float32[Array, "3"], Float32[Array, "5"]]:
  return x.sum(1), x.sum(0)


@function
def function3(
  x: Float32[Array, "3 5"],
  y: Float32[Array, "3 5"],
) -> Tuple[Float32[Array, "3"], Float32[Array, "5"]]:
  return x.sum(1), y.sum(0)


@function
def function4(
  x: Float32[Array, "3 5"],
  y: Float32[Array, "3 5"],
) -> Tuple[Tuple[Float32[Array, "3"], Float32[Array, ""]], Float32[Array, "5"]]:
  return (x.sum(1), y.sum()), x.sum(0)


class _TestGeneralArray(parameterized.TestCase):

  def test_spec(self):
    self.assertEqual(
      function0.shape,
      (Ret(Spec((), jnp.float32)), Arg(Spec((3, 5), jnp.float32)))
    )
    self.assertEqual(
      function1.shape,
      (Ret(Spec((3,), jnp.float32)), Arg(Spec((3, 5), jnp.float32)))
    )
    self.assertEqual(
      function2.shape, (
        Ret((Spec((3,), jnp.float32), Spec((5,), jnp.float32))
           ), Arg(Spec((3, 5), jnp.float32))
      )
    )

  @parameterized.parameters(
    (function0, 0, False),
    (function1, 0, False),
    (function2, 0, True),
    (function2, 0, False),
    (function3, 0, True),
    (function3, 0, False),
    (function3, 1, True),
    (function3, 1, False),
    (function3, (0, 1), True),
    (function3, (0, 1), False),
    (function4, 0, True),
    (function4, 0, False),
    (function4, 1, True),
    (function4, 1, False),
    (function4, (0, 1), True),
    (function4, (0, 1), False),
  )
  def test_jacobian(self, f, argnums, has_aux):
    jac_spec1 = jacobian_spec(f, argnums=argnums, has_aux=has_aux)
    jac = jax.jacrev(f, argnums=argnums, has_aux=has_aux)(*dummy_input(f))
    jac_spec2 = SpecTree.from_value(jac)
    self.assertEqual(jac_spec1, jac_spec2)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python3
# Copyright 2023 Garena Online Private Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp
from jax.tree_util import tree_map
from jaxtyping import Float32, Array
from typing import Tuple
import numpy as np
from absl.testing import absltest, parameterized
from autofd.general_array import (
  Spec,
  SpecTree,
  function,
  dummy_input,
  random_input,
  Grid,
)
import autofd.operators as o
from autofd.operators.operators import (
  split,
  compose,
  nabla,
  integrate,
  zip_functions,
  linear_transpose,
  linearize,
  permute_args,
  funcs,
  unary_op,
  binary_op,
  add,
  partial,
)


def _assert_tree_array_equal(a, b):
  tree_map(lambda x, y: np.testing.assert_array_equal(x, y), a, b)


def _assert_tree_array_almost_equal(a, b):
  tree_map(lambda x, y: np.testing.assert_array_almost_equal(x, y), a, b)


# Test the concat operator


@function
def concat_f(x: Float32[Array, "3 5"]) -> Float32[Array, "3"]:
  return jnp.sum(x, axis=-1)


@function
def concat_f1(x: Float32[Array, "3 5"]) -> Float32[Array, ""]:
  return jnp.max(x)


@function
def concat_f2(x: Float32[Array, "3 5"]) -> Float32[Array, "5"]:
  return x.min(axis=0)


@function
def concat_f3(
  x: Float32[Array, "3 5"],
) -> Tuple[Float32[Array, "3"], Float32[Array, ""], Float32[Array, "5"]]:
  return (x[:, 0], x[0, 0], x[0, :])


concat = lambda *fs: zip_functions(*fs, share_inputs=True)


class _TestConcatSplit(absltest.TestCase):

  def test_concat_split(self):
    f, f1, f2 = (concat_f, concat_f1, concat_f2)
    f012 = concat(f, f1, f2)
    x = jax.random.uniform(jax.random.PRNGKey(0), (3, 5))
    out = f012(x)
    expected = (f(x), f1(x), f2(x))
    _assert_tree_array_equal(out, expected)

    ff, ff1, ff2 = jax.linear_transpose(concat, f, f1, f2)(f012)
    out = (ff(x), ff1(x), ff2(x))
    _assert_tree_array_equal(out, expected)

    f, f1, f2 = split(f012)
    ff012 = jax.linear_transpose(split, f012)([f, f1, f2],)[0]
    out1, out2 = f012(x), ff012(x)
    _assert_tree_array_equal(out1, out2)

  def test_concat_split_vjp(self):
    """This test depends on modification on jax.
    specifically, we need jax._src.dtypes.dtype to support function input.
    """
    f, f1, f2, f3 = (concat_f, concat_f1, concat_f2, concat_f3)
    f012, vjp = jax.vjp(concat, f, f1, f2)
    ff1, ff2, ff3 = vjp(f3)
    x = jax.random.uniform(jax.random.PRNGKey(0), (3, 5))
    np.testing.assert_array_equal(ff1(x), x[:, 0])
    np.testing.assert_array_equal(ff2(x), x[0, 0])
    np.testing.assert_array_equal(ff3(x), x[0, :])

    a, b = jax.vjp(split, f012)
    (ff, ff1, ff2), vjp = jax.vjp(split, f012)
    ff012 = vjp([ff, ff1, ff2])[0]
    _assert_tree_array_equal(ff012(x), f012(x))

  def test_split_nabla(self):

    @function
    def f(a: Float32[Array, ""], b: Float32[Array, ""]) -> Float32[Array, ""]:
      return jnp.sin(a) + jnp.cos(b)

    def F1(f):
      fa1, fb1 = split(nabla(f, argnums=(0, 1)))
      return integrate(fa1 * fb1, argnums=(0, 1))

    def F2(f):
      fa2, fb2 = nabla(f, argnums=0), nabla(f, argnums=1)
      return integrate(fa2 * fb2, argnums=(0, 1))

    FD1 = jax.grad(F1)(f)
    FD2 = jax.grad(F2)(f)
    np.testing.assert_array_almost_equal(FD1(1., 2.), FD2(1., 2.))


@function
def linear_transpose_f1(
  x: Float32[Array, "3"],
  y: Float32[Array, "4"],
  z: Float32[Array, "5"],
) -> Tuple[Float32[Array, "3"], Float32[Array, "4"], Float32[Array, "5"]]:
  return (x * 2, y * 2, z * 2)


class _TestLinearTranspose(absltest.TestCase):

  def test_linear_transpose(self):
    f = linear_transpose_f1
    t = linear_transpose(f, argnums=(0, 1, 2))
    self.assertEqual(
      t.arg_spec, ((
        Spec((3,), "f"),
        Spec((4,), "f"),
        Spec((5,), "f"),
      ),)
    )
    self.assertEqual(SpecTree.from_value(t(*dummy_input(t))), t.ret_spec)

    t = linear_transpose(f, argnums=(0, 1, 2), unpack=True)
    self.assertEqual(t.shape, f.shape)
    self.assertEqual(SpecTree.from_value(t(*dummy_input(t))), t.ret_spec)

    t = linear_transpose(f, argnums=0)  # t (4,5,(3,4,5)) -> 3
    self.assertEqual(
      t.arg_spec, (
        Spec((4,), "f"),
        Spec((5,), "f"),
        (
          Spec((3,), "f"),
          Spec((4,), "f"),
          Spec((5,), "f"),
        ),
      )
    )
    self.assertEqual(t.ret_spec, Spec((3,), "f"))
    self.assertEqual(SpecTree.from_value(t(*dummy_input(t))), t.ret_spec)

    t = linear_transpose(f, argnums=(0,))
    self.assertEqual(
      t.arg_spec, (
        Spec((4,), "f"),
        Spec((5,), "f"),
        (
          Spec((3,), "f"),
          Spec((4,), "f"),
          Spec((5,), "f"),
        ),
      )
    )
    self.assertEqual(t.ret_spec, (Spec((3,), "f"),))
    self.assertEqual(SpecTree.from_value(t(*dummy_input(t))), t.ret_spec)

    t = linear_transpose(f, argnums=(0, 1))  # t (5,(3,4,5)) -> (3,4)
    self.assertEqual(
      t.arg_spec, (
        Spec((5,), "f"),
        (
          Spec((3,), "f"),
          Spec((4,), "f"),
          Spec((5,), "f"),
        ),
      )
    )
    self.assertEqual(t.ret_spec, (Spec((3,), "f"), Spec((4,), "f")))
    self.assertEqual(SpecTree.from_value(t(*dummy_input(t))), t.ret_spec)

  def test_linear_transpose_fd(self):
    f = linear_transpose_f1
    argnums = (0, 1)
    t = linear_transpose(f, argnums=argnums)
    ff = jax.linear_transpose(
      lambda f: linear_transpose(f, argnums=argnums), f
    )(t)[0]
    self.assertEqual(ff.shape, f.shape)
    args = random_input(jax.random.PRNGKey(0), ff)
    for i in argnums:
      np.testing.assert_array_equal(ff(*args)[i], f(*args)[i])


@function
def zip_f1(x: Float32[Array, "3 5"]) -> Float32[Array, "3"]:
  return jnp.sum(x, axis=-1)


@function
def zip_f2(x: Float32[Array, "3 5"]) -> Float32[Array, "5"]:
  return jnp.sum(x, axis=0)


class _TestZipFunction(absltest.TestCase):

  def test_zip_functions(self):
    a = zip_functions(zip_f1)

    @function
    def dummy(x: Float32[Array, "3 5"], /) -> Tuple[Float32[Array, "3"]]:
      pass

    @function
    def dummy1(
      x: Float32[Array, "3 5"]
    ) -> Tuple[Float32[Array, "3"], Float32[Array, "5"]]:
      pass

    @function
    def dummy2(
      x: Float32[Array, "3 5"], y: Float32[Array, "3 5"]
    ) -> Tuple[Float32[Array, "3"], Float32[Array, "5"]]:
      pass

    self.assertEqual(a.shape, dummy.shape)

    a = zip_functions(zip_f1, zip_f2, share_inputs=True)
    self.assertEqual(a.shape, dummy1.shape)
    a = zip_functions(zip_f1, zip_f2, share_inputs=False)
    self.assertEqual(a.shape, dummy2.shape)


# Test the compose operator


@function
def compose_g1(x: Float32[Array, "3 5"]) -> Float32[Array, "3"]:
  return jnp.sum(x, axis=-1)


@function
def compose_g2(x: Float32[Array, "3 5"]) -> Float32[Array, ""]:
  return jnp.max(x)


@function
def compose_f1(
  x: Float32[Array, "3"],
  y: Float32[Array, ""],
) -> Float32[Array, ""]:
  return jnp.min(x) + y


@function
def compose_g3(
  x: Float32[Array, "3 5"],
  y: Float32[Array, "3 5"],
) -> Float32[Array, "3"]:
  return jnp.sum(x, axis=-1) + jnp.max(y, axis=-1)


@function
def compose_g4(
  x: Float32[Array, "3 5"],
  y: Float32[Array, "3 5"],
) -> Tuple[Float32[Array, "3"], Float32[Array, ""]]:
  return (jnp.sum(x, axis=-1), jnp.max(y))


@function
def compose_f2(
  x: Tuple[Float32[Array, "3"], Float32[Array, ""]],
  y: Float32[Array, ""],
) -> Float32[Array, ""]:
  x, z = x
  return jnp.min(x) + y - z


class _TestCompose(absltest.TestCase):

  def setUp(self):
    super().setUp()

  def test_shape(self):
    pass

  def test_f1_call_g1_g2(self):
    g1, g2, f1 = (compose_g1, compose_g2, compose_f1)
    x = jax.random.normal(jax.random.PRNGKey(12), (3, 5))
    f = f1(g1, g2)
    y = f(x, x)
    self.assertEqual(y, x.sum(-1).min() + x.max())

  def test_f1_call_g1_g2_share_inputs(self):
    g1, g2, f1 = (compose_g1, compose_g2, compose_f1)
    x = jax.random.normal(jax.random.PRNGKey(12), (3, 5))
    f = f1(g1, g2, share_inputs=True)
    y = f(x)
    self.assertEqual(y, x.sum(-1).min() + x.max())

  def test_compose_f1_g1_g2(self):
    g1, g2, f1 = (compose_g1, compose_g2, compose_f1)
    x = jax.random.normal(jax.random.PRNGKey(12), (3, 5))
    f = compose(f1, g1, g2)
    y = f(x, x)
    self.assertEqual(y, x.sum(-1).min() + x.max())

  def test_compose_f1_g1_g2_share_inputs(self):
    g1, g2, f1 = (compose_g1, compose_g2, compose_f1)
    x = jax.random.normal(jax.random.PRNGKey(12), (3, 5))
    f = compose(f1, g1, g2, share_inputs=True)
    y = f(x)
    self.assertEqual(y, x.sum(-1).min() + x.max())

  def test_compose_f1_g3_g2(self):
    g2, g3, f1 = (compose_g2, compose_g3, compose_f1)
    x = jax.random.normal(jax.random.PRNGKey(12), (3, 5))
    y = jax.random.normal(jax.random.PRNGKey(42), (3, 5))
    f = compose(f1, g3, g2)
    z = f(x, y, x)
    self.assertEqual(z, (x.sum(-1) + y.max(-1)).min() + x.max())

  def test_compose_f2_g4_g2(self):
    g2, g4, f2 = (compose_g2, compose_g4, compose_f2)
    x = jax.random.normal(jax.random.PRNGKey(12), (3, 5))
    y = jax.random.normal(jax.random.PRNGKey(42), (3, 5))
    f = compose(f2, g4, g2)
    z = f(x, y, x)
    self.assertEqual(z, x.sum(-1).min() + x.max() - y.max())

  def test_compose_jvp_type(self):
    g2, g4, f2 = (compose_g2, compose_g4, compose_f2)
    primal_out, tangent_out = jax.jvp(compose, (f2, g4, g2), (f2, g4, g2))
    self.assertEqual(tangent_out.shape, primal_out.shape)

  def test_compose_vjp(self):

    @function
    def f(x: Float32[Array, ""]) -> Float32[Array, ""]:
      return x * 2

    @function
    def g(x: Float32[Array, ""]) -> Float32[Array, ""]:
      return x**2

    @function
    def o(x: Float32[Array, ""]) -> Float32[Array, ""]:
      return x**2 * 2

    def comp_f(g):
      return compose(f, g)

    def vjp_comp_f(g, o):
      out, vjp = jax.vjp(comp_f, g)
      return vjp(o)[0]

    out, vjp = jax.vjp(comp_f, g)
    gg = vjp(out)[0]
    gg(*dummy_input(gg))

  def test_higher_order_el(self):

    def euler_lagrange(I, f):

      def I_args(*f_args):
        return f(*f_args), jax.jacfwd(f)(*f_args), jax.hessian(f)(*f_args)

      def el(*f_args):
        return (
          jax.grad(I, argnums=0)(*I_args(*f_args)) -
          jax.grad(lambda *f_args: jax.grad(I, argnums=1)
                   (*I_args(*f_args)))(*f_args) +
          jax.hessian(lambda *f_args: jax.grad(I, argnums=2)
                      (*I_args(*f_args)))(*f_args)
        )

      return el

    def I(f, df, ddf):
      return jnp.sum(f) + jnp.sum(df) + jnp.sum(ddf)

    def F(f):
      df = nabla(f)
      ddf = nabla(df)
      If = compose(I, f, df, ddf, share_inputs=True)
      return integrate(If, argnums=0)

    @function
    def f(x: Float32[Array, ""]) -> Float32[Array, ""]:
      return jnp.sin(x) + 0.3 * jnp.cos(x)

    fd1 = jax.grad(F)(f)
    fd2 = euler_lagrange(I, f)
    x = jax.random.normal(jax.random.PRNGKey(12), (10,))
    np.testing.assert_array_equal(jax.vmap(fd1)(x), jax.vmap(fd2)(x))

  def test_euler_lagrange(self):

    @function
    def multiply_f_nablaf(f: Float32[Array, ""], nablaf: Float32[Array, ""],
                          /) -> Float32[Array, ""]:
      return jnp.multiply(f, nablaf)

    @function
    def sum_square_f_nablaf(
      f: Float32[Array, ""], nablaf: Float32[Array, ""], /
    ) -> Float32[Array, ""]:
      return jnp.add(f**2, nablaf**2)

    @function
    def square_f(f: Float32[Array, ""], nablaf: Float32[Array, ""],
                 /) -> Float32[Array, ""]:
      return nablaf**2

    integrand = sum_square_f_nablaf

    def I(f):
      return integrate(f**2 + nabla(f)**2)

    # do this again with euler lagrange equation
    def DIDf(f):

      def func(x):
        # D I / D f = p I / p f - d/dx ( p I / p f')
        pIpf = jax.grad(integrand, argnums=0)(f(x), nabla(f)(x))
        ddx_pIpnablaf = jax.grad(
          lambda x: jax.grad(integrand, argnums=1)(f(x), nabla(f)(x))
        )(
          x
        )
        return pIpf - ddx_pIpnablaf

      return func

    # evaluate the functional on sin function
    @function
    def sin(x: Float32[Array, ""], /) -> Float32[Array, ""]:
      return jnp.sin(x)

    sin.support = (jnp.array(0.),)
    fd1 = jax.grad(I)(sin)
    fd2 = DIDf(sin)
    x = jax.random.normal(jax.random.PRNGKey(12), (10,))
    np.testing.assert_array_equal(jax.vmap(fd1)(x), jax.vmap(fd2)(x))


# Test add
@function
def add_f1(
  x: Float32[Array, "3 5"],
  y: Float32[Array, "3 5"],
) -> Tuple[Float32[Array, "3"], Float32[Array, ""]]:
  return jnp.min(x, axis=-1), jnp.min(y)


@function
def add_f2(
  x: Float32[Array, "3 5"],
  y: Float32[Array, "3 5"],
) -> Tuple[Float32[Array, "3"], Float32[Array, ""]]:
  return jnp.sum(x, axis=-1), jnp.max(y)


class _TestBinaryFunctions(absltest.TestCase):

  def test_add(self):
    f1, f2 = (add_f1, add_f2)
    f12 = add(f1, f2)
    x, y = random_input(jax.random.PRNGKey(12), f12)
    _assert_tree_array_equal(f12(x, y), tree_map(jnp.add, f1(x, y), f2(x, y)))

  def test_sub(self):
    f1, f2 = (add_f1, add_f2)
    f12 = f1 - f2
    x, y = random_input(jax.random.PRNGKey(12), f12)
    _assert_tree_array_equal(
      f12(x, y), tree_map(jnp.subtract, f1(x, y), f2(x, y))
    )


# Test the nabla operator


@function
def nabla_f(x: Float32[Array, "3 5"]) -> Float32[Array, "3"]:
  return jnp.sum(jnp.sin(x * 2), axis=-1)


@function
def nabla_f1(x: Float32[Array, "3 5"]) -> Float32[Array, ""]:
  return jnp.sum(jnp.sin(x * 2))


@function
def nabla_f2(
  x: Float32[Array, "3 5"]
) -> Tuple[Float32[Array, ""], Float32[Array, "3"]]:
  return jnp.sum(jnp.sin(x * 2)), jnp.sin(x)[:, 0]


@function
def nabla_f3(x: Float32[Array, ""]) -> Float32[Array, ""]:
  return jnp.sin(x)


nabla_fs = (nabla_f, nabla_f1, nabla_f2, nabla_f3)


class _TestNabla(parameterized.TestCase):

  @parameterized.parameters(*((f, i) for f in nabla_fs for i in range(1, 2)))
  def test_nabla(self, f, recursion=1):
    jac_fn1, jac_fn2 = f, f
    for _ in range(recursion):
      jac_fn1 = nabla(jac_fn1)
      jac_fn2 = jax.jacrev(jac_fn2)

    args = random_input(jax.random.PRNGKey(42), f)
    jac1, jac2 = jac_fn1(*args), jac_fn2(*args)
    tree_map(np.testing.assert_array_equal, jac1, jac2)

  def test_nabla_nabla(self):

    @function
    def f(x: Float32[Array, "3"]) -> Float32[Array, ""]:
      return jnp.sum(jnp.sin(x))

    args = random_input(jax.random.PRNGKey(42), f)

    def F(f):
      dfdf = compose(lambda x: jnp.sum(x), nabla(f)**2)
      return integrate(dfdf)

    def FF(f):
      ddfddf = compose(lambda x: jnp.sum(x), nabla(nabla(f))**2)
      return integrate(ddfddf)

    fd = jax.grad(F)(f)
    fd1 = compose(lambda x: -jnp.trace(x), nabla(2 * nabla(f)))
    np.testing.assert_array_equal(fd(*args), fd1(*args))

    ffd = jax.grad(FF)(f)
    trace = lambda x: -jnp.trace(x, axis1=-2, axis2=-1)
    ffd1 = compose(trace, nabla(compose(trace, nabla(2 * nabla(nabla(f))))))
    np.testing.assert_array_equal(ffd(*args), ffd1(*args))

  @parameterized.parameters(*((f,) for f in nabla_fs))
  def test_transpose(self, f):
    jacf = nabla(f)
    ff = jax.linear_transpose(nabla, f)(jacf)[0]
    self.assertEqual(ff.arg_spec, f.arg_spec)
    self.assertEqual(ff.ret_spec, f.ret_spec)
    args = random_input(jax.random.PRNGKey(42), ff)
    out = ff(*args)
    self.assertEqual(SpecTree.from_value(out), ff.ret_spec)

  @parameterized.parameters(*((f,) for f in nabla_fs))
  def test_functional_derivative(self, f):
    nabla_f = nabla(f)
    _, nabla_vjp = jax.vjp(nabla, f)
    g = nabla_vjp(nabla_f)[0]
    self.assertEqual(g.arg_spec, f.arg_spec)

  def test_known_functions(self):

    @function
    def sin(x: Float32[Array, ""]) -> Float32[Array, ""]:
      return jnp.sin(x)

    nabla_sin = nabla(sin)
    xs = jax.random.normal(jax.random.PRNGKey(0), (100,))
    for x in xs:
      np.testing.assert_allclose(nabla_sin(x), jnp.cos(x))

    nabla_vjp = jax.vjp(nabla, sin)[1]
    nabla_t_sin = nabla_vjp(sin)[0]
    for x in xs:
      np.testing.assert_allclose(nabla_t_sin(x), -jnp.cos(x))


class _TestLinearize(absltest.TestCase):

  def test_linearize(self):

    @function
    def f(
      a: Float32[Array, "3 5"],
      b: Float32[Array, "3 5"],
    ) -> Float32[Array, "3 5"]:
      return a**2 + jnp.sin(b)

    x, y = (
      jax.random.normal(jax.random.PRNGKey(0), (3, 5)),
      jax.random.normal(jax.random.PRNGKey(1), (3, 5)),
    )
    x_dot, y_dot = (
      jax.random.normal(jax.random.PRNGKey(2), (3, 5)),
      jax.random.normal(jax.random.PRNGKey(3), (3, 5)),
    )
    out0 = linearize(f, argnums=(0,))(x, y, x_dot)
    expect0 = 2 * x * x_dot
    out1 = linearize(f, argnums=(1,))(x, y, y_dot)
    expect1 = jnp.cos(y) * y_dot
    out2 = linearize(f, argnums=(0, 1))(x, y, x_dot, y_dot)
    expect2 = 2 * x * x_dot + jnp.cos(y) * y_dot
    np.testing.assert_array_equal(out0, expect0)
    np.testing.assert_array_equal(out1, expect1)
    np.testing.assert_array_equal(out2, expect2)


@function
def overload_f1(
  a: Float32[Array, "3 5"],
  b: Float32[Array, "3 5"],
) -> Float32[Array, "3 5"]:
  return a**2 + jnp.sin(b)


@function
def overload_f2(
  a: Float32[Array, "3 5"],
  b: Float32[Array, "3 5"],
) -> Tuple[Float32[Array, "3 5"], Float32[Array, "3 5"]]:
  return (a**2, jnp.sin(b))


@function
def overload_f3(
  a: Float32[Array, "3 5"],
  b: Float32[Array, "3 5"],
) -> Tuple[Float32[Array, "5"], Float32[Array, "5"]]:
  return ((a**2).sum(0), jnp.sin(b).sum(0))


class _TestOperatorOverload(parameterized.TestCase):

  @parameterized.parameters(overload_f1, overload_f2, overload_f3)
  def test_add(self, f):
    f2 = f + f
    args = random_input(jax.random.PRNGKey(0), f2)
    _assert_tree_array_equal(f2(*args), tree_map(lambda x: 2 * x, f(*args)))
    fp1 = f + 1
    _assert_tree_array_equal(fp1(*args), tree_map(lambda x: x + 1, f(*args)))

  @parameterized.parameters(overload_f1, overload_f2, overload_f3)
  def test_sub(self, f):
    g = function(f)
    f2 = g + g
    f2subf = f2 - f
    fsubf2 = f - f2
    args = random_input(jax.random.PRNGKey(0), f2subf)
    np.testing.assert_array_equal(f2subf(*args), f(*args))
    np.testing.assert_array_equal(
      fsubf2(*args), tree_map(lambda x: -x, f(*args))
    )

  @parameterized.parameters(overload_f1, overload_f2, overload_f3)
  def test_neg(self, f):
    g = function(f)
    args = random_input(jax.random.PRNGKey(0), g)
    np.testing.assert_array_equal((-g)(*args), tree_map(lambda x: -x, g(*args)))

  @parameterized.parameters(overload_f1, overload_f2, overload_f3)
  def test_pow(self, f):
    f = overload_f1
    g = function(f)
    args = random_input(jax.random.PRNGKey(0), g)
    _assert_tree_array_almost_equal(
      jax.jit(g**3)(*args), tree_map(lambda x: x**3, g(*args))
    )

  @parameterized.parameters(overload_f1, overload_f2, overload_f3)
  def test_mul(self, f):
    g = function(f)
    args = random_input(jax.random.PRNGKey(0), g)
    np.testing.assert_array_equal(
      (g * f)(*args), tree_map(lambda x: x**2, f(*args))
    )
    np.testing.assert_array_equal(
      (f * g)(*args), tree_map(lambda x: x**2, f(*args))
    )

  @parameterized.parameters(overload_f1, overload_f2, overload_f3)
  def test_div(self, f):
    g = function(f)
    args = random_input(jax.random.PRNGKey(0), g)
    _assert_tree_array_almost_equal(
      jax.jit(g / f)(*args), tree_map(jnp.ones_like, f(*args))
    )
    _assert_tree_array_almost_equal(
      (f / g)(*args), tree_map(jnp.ones_like, f(*args))
    )


def _nary_fns(num_args=1):
  for k, fns in funcs.items():
    for name, nary in fns.items():
      if (
        (nary is unary_op and num_args == 1) or
        (nary is binary_op and num_args == 2)
      ):
        yield getattr(getattr(jax, k), name), getattr(getattr(o, k), name)


class _TestUnary(parameterized.TestCase):

  @parameterized.parameters(
    (func, operator, f)
    for func, operator in _nary_fns(1)
    for f in (overload_f1, overload_f2)
  )
  def test_unary_compose(self, func, operator, f):
    u_f = operator(f)
    args = random_input(jax.random.PRNGKey(0), u_f)
    out0 = u_f(*args)
    out1 = tree_map(func, f(*args))
    _assert_tree_array_almost_equal(out0, out1)

  @parameterized.parameters(_nary_fns(2))
  def test_binary_compose(self, func, operator):
    f1, f2 = overload_f1, overload_f2
    b_f = operator(f1, f2)
    args = random_input(jax.random.PRNGKey(0), b_f)
    out0 = b_f(*args)
    out1 = tree_map(lambda y: func(f1(*args), y), f2(*args))
    _assert_tree_array_equal(out0, out1)

  @parameterized.parameters(overload_f1, overload_f2)
  def test_unary_with_params(self, f):
    sf = o.numpy.sum(f, axis=0)
    out0 = sf(*random_input(jax.random.PRNGKey(0), sf))
    out1 = tree_map(
      lambda x: x.sum(0), f(*random_input(jax.random.PRNGKey(0), f))
    )
    _assert_tree_array_equal(out0, out1)


class _TestPermuteArgs(absltest.TestCase):

  def test_permute_args(self):

    @function
    def f(
      a: Float32[Array, "5 4"],
      b: Float32[Array, "4 3"],
      c: Float32[Array, "3 2"],
    ) -> Float32[Array, "5 2"]:
      return a @ b @ c

    args = random_input(jax.random.PRNGKey(0), f)
    np.testing.assert_array_equal(
      f(*args),
      permute_args(f, perm=(2, 1, 0))(*args[::-1])
    )
    f201 = permute_args(f, perm=(2, 0, 1))
    f201 = jax.linear_transpose(
      lambda f: permute_args(f, perm=(1, 2, 0)), f201
    )(f)[0]
    np.testing.assert_array_equal(f201(args[2], args[0], args[1]), f(*args))


class _TestBroadcast(absltest.TestCase):

  def test_broadcast(self):

    @function
    def f(
      a: Float32[Array, "5 4"],
      b: Float32[Array, "4 3"],
    ) -> Float32[Array, "5 3"]:
      return a @ b

    @function
    def target(
      a: Float32[Array, "5 4"],
      b: Float32[Array, "4 3"],
      c: Float32[Array, "3 3"],
    ) -> Float32[Array, "5 3"]:
      return a @ b @ c

    a, b, c = random_input(jax.random.PRNGKey(0), target)
    fb = o.broadcast_arg(f, target, argnums=(0, 1))
    o1 = fb(a, b, c)
    o2 = f(a, b)
    np.testing.assert_array_equal(o1, o2)


class _TestPartial(absltest.TestCase):

  def test_partial(self):

    @function
    def f(
      a: Float32[Array, "5 4"],
      b: Float32[Array, "4 3"],
      c: Float32[Array, "3 2"],
    ) -> Float32[Array, "5 2"]:
      return a @ b @ c

    @function
    def tbc(
      b: Float32[Array, "4 3"],
      c: Float32[Array, "3 2"],
    ) -> Float32[Array, "5 2"]:
      return jnp.ones((5, 4)) @ b @ c

    @function
    def tb(b: Float32[Array, "4 3"],) -> Float32[Array, "5 2"]:
      return jnp.ones((5, 4)) @ b @ jnp.ones((3, 2))

    a, b, c = random_input(jax.random.PRNGKey(0), f)
    fa = partial(f, args=(a,), argnums=0)
    np.testing.assert_array_almost_equal(fa(b, c), f(a, b, c))
    fab = partial(f, args=(a, b), argnums=(0, 1))
    np.testing.assert_array_almost_equal(fab(c), f(a, b, c))
    tbc.support = (b, c)
    ta = jax.linear_transpose(
      lambda a: partial(f, args=(a,), argnums=0),
      a,
    )(tbc)[0]
    np.testing.assert_array_almost_equal(
      ta,
      integrate(tbc, argnums=(0, 1)) @ c.T @ b.T
    )
    tb.support = (b,)
    ta = jax.linear_transpose(
      lambda a: partial(f, args=(a, c), argnums=(0, 2)),
      a,
    )(tb)[0]
    np.testing.assert_array_almost_equal(ta, integrate(tb) @ c.T @ b.T)

  def _test_partial_gradient(self):

    @function
    def f(
      a: Float32[Array, "5 4"],
      b: Float32[Array, "4 3"],
      c: Float32[Array, "3 2"],
    ) -> Float32[Array, "5 2"]:
      return a @ b @ c

    a, b, c = random_input(jax.random.PRNGKey(0), f)

    def func1(a, b, c):
      fab = partial(f, args=(a, b, c), argnums=(0, 1, 2))
      return integrate(compose(jnp.sum, fab))

    fd1 = jax.grad(func1, argnums=(0, 1))(a, b, c)

    def func2(a, b, c):
      return jnp.sum(f(a, b, c))

    fd2 = jax.grad(func2, argnums=(0, 1))(a, b, c)
    _assert_tree_array_equal(fd1, fd2)

  def test_partial_jit(self):

    @function
    def f(a: Float32[Array, ""], b: Float32[Array, ""]) -> Float32[Array, ""]:
      return a + b

    def F(f):
      return o.integrate(o.numpy.sin(f))

    @jax.jit
    def loss(a, b):
      with jax.ensure_compile_time_eval():
        g = o.partial(f, args=(a,), argnums=(0,))
        fd = jax.grad(F)(g)
      return fd(b)

    l = loss(1., 2.)
    print(l)

    g = jax.grad(loss)(1., 2.)
    print(g)


class _TestBraKet(absltest.TestCase):

  def test_braket(self):

    @function
    def lfun(a: Float32[Array, ""]) -> Float32[Array, ""]:
      return jnp.sin(a)

    @function
    def rfun(a: Float32[Array, ""]) -> Float32[Array, ""]:
      return jnp.cos(a)

    lfun.grid = Grid(
      nodes=(jnp.arange(10, dtype=jnp.float32),),
      weights=(jnp.ones(10, dtype=jnp.float32) / 10,),
    )
    rfun.grid = lfun.grid
    lr1 = o.braket(lfun, rfun)
    lr2 = (jnp.sin(jnp.arange(10)) * jnp.cos(jnp.arange(10))).mean()
    np.testing.assert_array_almost_equal(lr1, lr2)

  def test_braket_grad(self):

    @function
    def lfun(a: Float32[Array, ""]) -> Float32[Array, ""]:
      return jnp.sin(a)

    lfun.grid = Grid(
      nodes=(jnp.arange(10, dtype=jnp.float32),),
      weights=(jnp.ones(10, dtype=jnp.float32) / 10,),
    )

    def F(f):
      return o.braket(f, f, real=True)

    dFdf = jax.grad(F)(lfun)
    a, = random_input(jax.random.PRNGKey(0), lfun)
    out1 = dFdf(a)
    out2 = 2 * jnp.sin(a)
    np.testing.assert_array_almost_equal(out1, out2)


if __name__ == "__main__":
  absltest.main()
